import logging
from math import ceil
from typing import Optional

from django.db.models import Q
from django.db.models.deletion import Collector
from django.db.models.signals import pre_delete
from django_rq import job

from arkindex.dataimport.models import DataImportElement
from arkindex.documents.indexer import Indexer
from arkindex.documents.models import (
    Classification,
    Corpus,
    Element,
    ElementPath,
    Entity,
    EntityLink,
    MetaData,
    Selection,
    Transcription,
    TranscriptionEntity,
)
from arkindex.documents.signals import pre_delete_handler
from arkindex.project.tools import disconnect_signal

logger = logging.getLogger(__name__)


def _delete_queryset(queryset, batch_size=1000):
    """
    Helper to delete large querysets with as little SQL queries and memory footprint as possible.
    """
    count = queryset.count()
    logger.info('Deleting {} {}'.format(count, queryset.model.__name__))

    if not count:
        return

    if Collector(using=queryset.db).can_fast_delete(queryset.all()):
        # If a single DELETE statement can be used,
        # bypass both the batched deletion and Django's related objects checks.
        logger.debug('Using single-query deletion')
        queryset._raw_delete(using=queryset.db)
        return

    if count <= batch_size:
        # If there is a single batch, just delete.
        queryset.delete()
        return

    for i in range(ceil(count / batch_size)):
        logger.debug('Deleting batch {}'.format(i + 1))
        # Deleting a slice is not allowed;
        # we use a sliced subquery instead and still delete in a single query.
        # DELETE FROM … WHERE id IN (SELECT id FROM … LIMIT [batch_size])
        ids = queryset[:batch_size].values('id')
        queryset.model.objects.filter(id__in=ids).delete()


@job
def reindex_start(corpus_id: Optional[str] = None,
                  element_id: Optional[str] = None,
                  entity_id: Optional[str] = None,
                  transcriptions: bool = True,
                  elements: bool = True,
                  entities: bool = True,
                  drop: bool = False) -> None:
    indexer = Indexer()

    if drop:
        if transcriptions:
            indexer.drop_index(Transcription.es_document)
        if elements:
            indexer.drop_index(Element.es_document)
        if entities:
            indexer.drop_index(Entity.es_document)
        indexer.setup()

    if entity_id:
        # Pick entity only
        transcriptions_queryset = Transcription.objects.none()
        elements_queryset = Element.objects.none()
        entities_queryset = Entity.objects.filter(id=entity_id)
    elif element_id or corpus_id:
        if element_id:
            # Pick this element, and all its children
            elements_queryset = Element.objects.filter(Q(id=element_id) | Q(paths__path__contains=[element_id]))
        else:
            # Pick all elements in the corpus
            elements_queryset = Element.objects.filter(corpus_id=corpus_id)

        transcriptions_queryset = Transcription.objects.filter(element__in=elements_queryset)
        entities_queryset = Entity.objects.filter(
            Q(metadatas__element__in=elements_queryset)
            | Q(transcriptions__element__in=elements_queryset)
        )
    else:
        transcriptions_queryset = Transcription.objects.all()
        elements_queryset = Element.objects.all()
        entities_queryset = Entity.objects.all()

    if transcriptions:
        indexer.run_index(transcriptions_queryset.select_related('element'), bulk_size=400)
    if elements:
        indexer.run_index(
            elements_queryset.select_related('type').prefetch_related('metadatas', 'transcriptions'),
            bulk_size=100,
        )
    if entities:
        indexer.run_index(entities_queryset, bulk_size=400)


@job
def ml_results_delete(corpus_id: Optional[str] = None,
                      element_id: Optional[str] = None,
                      batch_size: int = 1000) -> None:
    assert corpus_id or element_id, 'Missing element or corpus IDs'

    if element_id:
        logger.info('Deleting ML results on element {}'.format(element_id))
        element = Element.objects.get(id=element_id)
        if element.type.folder:
            # The folder AND its children
            elements = Element.objects.filter(id=element_id).values('id').union(
                # Disable ordering here because we do not need it and it adds an extra column,
                # causing the UNION to fail
                Element.objects.get_descending(element_id).order_by().values('id')
            )
        else:
            elements = [element]

        if not corpus_id:
            # The corpus ID is still used in some deletions; deduce it from the element.
            corpus_id = Element.objects.get(id=element_id).corpus_id
    elif corpus_id:
        logger.info('Deleting ML results on corpus {}'.format(corpus_id))
        elements = Element.objects.filter(corpus_id=corpus_id)

    # Simple deletions for classifications and transcriptions.
    _delete_queryset(Classification.objects.filter(element__in=elements).exclude(source__slug='manual'), batch_size)
    _delete_queryset(Transcription.objects.filter(element__in=elements).exclude(source__slug='manual'), batch_size)

    # Entity deletion is complex: they can be linked to different elements both on transcriptions and metadata.
    # Metadata are not considered ML results so we need to keep them: update them to unlink entities.
    logger.info('Updating element metadata')
    MetaData.objects.filter(element__in=elements).update(entity_id=None)

    # We removed transcriptions earlier, which implies removing the links with entities.
    # All is left is to remove 'lonely' entities.
    # Note: __isnull's implementation will fetch all element IDs into a list before deleting—use batches!
    _delete_queryset(Entity.objects.filter(
        corpus_id=corpus_id,
        metadatas__isnull=True,
        transcriptions__isnull=True,
    ), batch_size)


@job
def corpus_delete(corpus_id: str, batch_size: int = 1000) -> None:
    corpus = Corpus.objects.get(id=corpus_id)

    logger.info('Deleting {!r}'.format(corpus))

    # Delete all related objects, bypassing RESTRICTs deletion rules
    # and Django's way of loading everything into memory before deleting.
    querysets = [
        DataImportElement.objects.filter(
            Q(dataimport__in=corpus.imports.all())
            | Q(element__in=corpus.elements.all())
        ),
        corpus.imports.all(),
        corpus.files.all(),
        EntityLink.objects.filter(role__corpus_id=corpus_id),
        TranscriptionEntity.objects.filter(
            Q(entity__in=corpus.entities.all())
            | Q(transcription__in=Transcription.objects.filter(element__in=corpus.elements.all()))
        ),
        corpus.entities.all(),
        corpus.roles.all(),
        corpus.allowed_metadatas.all(),
        MetaData.objects.filter(element__in=corpus.elements.all()),
        Classification.objects.filter(element__in=corpus.elements.all()),
        corpus.ml_classes.all(),
        Transcription.objects.filter(element__in=corpus.elements.all()),
        ElementPath.objects.filter(element__in=corpus.elements.all()),
        Selection.objects.filter(element__in=corpus.elements.all()),
        corpus.elements.all(),
        corpus.types.all(),
    ]

    with disconnect_signal(pre_delete, sender=Element, receiver=pre_delete_handler):
        for queryset in querysets:
            _delete_queryset(queryset, batch_size)

    corpus.delete()
    logger.info('Deleted corpus {}'.format(corpus_id))
