from django.db.models.signals import pre_delete

from arkindex.dataimport.models import Repository, RepositoryType
from arkindex.documents.models import Corpus, DataSource, Element, MLClass
from arkindex.documents.tasks import corpus_delete
from arkindex.project.tests import FixtureTestCase
from arkindex_common.enums import DataImportMode, MetaType, TranscriptionType
from arkindex_common.ml_tool import MLToolType


class TestDeleteCorpus(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.img1 = cls.imgsrv.images.get(path='img1')
        cls.zone = cls.img1.zones.create(polygon=[(0, 0), (10, 0), (10, 10), (0, 10), (0, 0)])
        cls.repo = Repository.objects.create(
            url='http://lol.git',
            type=RepositoryType.IIIF,
            hook_token='h00k',
            provider_name='gitlab',
        )
        cls.corpus2 = Corpus.objects.create(name='Other corpus', repository=cls.repo)
        cls.rev = cls.repo.revisions.create(
            hash='42',
            message='oh',
            author='me',
        )
        cls.di = cls.rev.dataimports.create(
            creator=cls.user,
            corpus=cls.corpus2,
            mode=DataImportMode.Repository,
        )
        cls.df = cls.di.files.create(
            name='a.txt',
            size=1234,
            content_type='text/plain',
            corpus=cls.corpus2,
        )
        cls.vol = cls.corpus2.elements.create(
            type=cls.corpus2.types.create(slug='volume', folder=True),
            name='A volume',
        )
        cls.vol.metadatas.create(
            type=MetaType.Text,
            name='meta',
            value='data',
        )
        cls.page = Element.objects.create(
            corpus=cls.corpus2,
            type=cls.corpus2.types.create(slug='page'),
            name='A page',
        )
        cls.page.classifications.create(
            source=DataSource.objects.create(
                type=MLToolType.Classifier,
                slug='classeur',
                revision='Early Access',
                internal=False,
            ),
            ml_class=MLClass.objects.create(
                name='klass',
                corpus=cls.corpus2,
            ),
            confidence=0.5,
        )
        cls.page.transcriptions.create(
            source=DataSource.objects.create(
                type=MLToolType.Recognizer,
                slug='reco',
                revision='-1',
                internal=False,
            ),
            type=TranscriptionType.Word,
            text='hi',
            score=0.75,
        )
        cls.page.add_parent(cls.vol)

    def test_run(self):
        receivers = pre_delete.receivers

        self.corpus.repository = self.repo
        self.corpus.save()

        with self.assertNumQueries(47):
            corpus_delete(self.corpus.id)

        # Ensure the command restores the signal receivers
        self.assertEqual(pre_delete.receivers, receivers)

        self.assertEqual(Corpus.objects.count(), 1)
        with self.assertRaises(Corpus.DoesNotExist):
            self.corpus.refresh_from_db()
        self.corpus2.refresh_from_db()
        self.assertEqual(self.corpus2.types.count(), 2)
        self.assertEqual(self.corpus2.elements.count(), 2)
        self.assertEqual(Element.objects.count(), 2)
        self.assertIn(self.corpus.repository, Repository.objects.all())

        self.repo.refresh_from_db()
        self.rev.refresh_from_db()
        self.di.refresh_from_db()
        self.df.refresh_from_db()
        self.vol.refresh_from_db()
        self.page.refresh_from_db()

        self.assertTrue(self.repo.revisions.filter(id=self.rev.id).exists())
        self.assertEqual(self.di.revision, self.rev)
        self.assertEqual(self.di.files.get(), self.df)
        self.assertTrue(Element.objects.get_descending(self.vol.id).filter(id=self.page.id).exists())

        md = self.vol.metadatas.get()
        self.assertEqual(md.name, 'meta')
        self.assertEqual(md.value, 'data')

        cl = self.page.classifications.get()
        self.assertEqual(cl.source.slug, 'classeur')
        self.assertEqual(cl.ml_class.name, 'klass')
        self.assertEqual(cl.confidence, 0.5)

        ts = self.page.transcriptions.get()
        self.assertEqual(ts.source.slug, 'reco')
        self.assertEqual(ts.type, TranscriptionType.Word)
        self.assertEqual(ts.text, 'hi')
        self.assertEqual(ts.score, 0.75)
