"""
Helper methods to trigger tasks in asynchronous workers
"""
from typing import Union
from uuid import UUID

from django.conf import settings

from arkindex.documents import tasks
from arkindex.documents.models import Corpus, Element, Entity


def reindex_start(*,
                  corpus: Union[Corpus, UUID, str] = None,
                  element: Union[Element, UUID, str] = None,
                  entity: Union[Entity, UUID, str] = None,
                  transcriptions: bool = False,
                  elements: bool = False,
                  entities: bool = False,
                  drop: bool = False) -> None:
    """
    Reindex elements into ElasticSearch.

    One of `corpus`, `element` or `entity` may be specified.
    Otherwise all elements will be picked.
    If `element` is specified, its children will be included.

    `transcriptions`, `elements` and `entities` booleans represents Elasticsearch corresponding indexes.
    Update will be run on indexes set to True only.

    `drop` will REMOVE all indexes entirely and recreate them before indexing.
    This is only allowed when reindexing all elements.
    """
    if not settings.ARKINDEX_FEATURES['search']:
        return

    element_id = None
    corpus_id = None
    entity_id = None
    if isinstance(element, Element):
        element_id = str(element.id)
    elif element:
        element_id = str(element)

    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
    elif corpus:
        corpus_id = str(corpus)

    if isinstance(entity, Entity):
        entity_id = str(entity.id)
    elif entity:
        entity_id = str(entity)

    tasks.reindex_start.delay(
        element_id=element_id,
        corpus_id=corpus_id,
        entity_id=entity_id,
        transcriptions=transcriptions,
        elements=elements,
        entities=entities,
        drop=drop,
    )


def ml_results_delete(*,
                      element: Union[Element, UUID, str] = None,
                      corpus: Union[Corpus, UUID, str] = None,
                      batch_size: int = 1000) -> None:
    """
    Delete all ML results from all sources on a corpus
    or an element and its *direct* (non-recursive) children.
    """
    element_id = None
    corpus_id = None
    if isinstance(element, Element):
        element_id = str(element.id)
    elif element:
        element_id = str(element)

    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
    elif corpus:
        corpus_id = str(corpus)

    assert element_id or corpus_id, 'Missing element or corpus ID'

    job = tasks.ml_results_delete.delay(corpus_id=corpus_id, element_id=element_id, batch_size=batch_size)
    if settings.ARKINDEX_FEATURES['search']:
        # Trigger a reindex afterwards to cleanup the deleted results
        tasks.reindex_start.delay(corpus_id=corpus_id, element_id=element_id, depends_on=job)


def corpus_delete(corpus: Union[Corpus, UUID, str],
                  *,
                  batch_size: int = 1000) -> None:
    """
    Delete a whole corpus without killing a server by removing all related
    models first and lowering the amount of foreign keys Django has to handle.
    """
    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
    else:
        corpus_id = str(corpus)

    tasks.corpus_delete.delay(corpus_id=corpus_id, batch_size=batch_size)
